-- @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
-- @@@@@@@@@  PACKAGE->TBL_PACKAGE_DETAILS TBL_PACKAGE_DETAILS TBL_PACKAGE_DETAILS  @@@@@@@@@
-- @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
------------------------------------------------------------------------------------------
            CREATE OR REPLACE FUNCTION fn_get_all_package_details() RETURNS SETOF package.tbl_package_details AS $$
                DECLARE

                BEGIN
                    RETURN QUERY SELECT * FROM package.tbl_package_details WHERE int_sys_action_id != -1;  
                END;
            $$ LANGUAGE plpgsql;
        

------------------------------------------------------------------------------------------
            CREATE OR REPLACE FUNCTION fn_get_all_package_details_with_action_user() 
                            RETURNS TABLE (pk_package_details_id BIGINT,
								fk_package_master_id BIGINT,
								sin_type SMALLINT,

								fk_commission_ac_id BIGINT,
								fk_sup_fee_ac_id BIGINT,
								fk_supplier_ac_id BIGINT,
								fk_service_fee_ac_id BIGINT,
								fk_discount_ac_id BIGINT,

								txt_particulars TEXT,

								dbl_fare DOUBLE PRECISION,
								dbl_tax DOUBLE PRECISION,
								dbl_commission DOUBLE PRECISION,
								dbl_sup_fee DOUBLE PRECISION,
								dbl_cost DOUBLE PRECISION,
								dbl_service_fee DOUBLE PRECISION,
								dbl_discount DOUBLE PRECISION,
								dbl_price DOUBLE PRECISION) AS $$
                DECLARE

                BEGIN
                    RETURN QUERY SELECT mt.pk_package_details_id,
					mt.fk_package_master_id,
					mt.sin_type,

					mt.fk_commission_ac_id,
					mt.fk_sup_fee_ac_id,
					mt.fk_supplier_ac_id,
					mt.fk_service_fee_ac_id,
					mt.fk_discount_ac_id,

					mt.txt_particulars,

					mt.dbl_fare,
					mt.dbl_tax,
					mt.dbl_commission,
					mt.dbl_sup_fee,
					mt.dbl_cost,
					mt.dbl_service_fee,
					mt.dbl_discount,
					mt.dbl_price
                        FROM package.tbl_package_details mt 
                            LEFT JOIN authentication.tbl_user u1
                                ON mt.fk_created_user_id =  u1.pk_user_id
                            LEFT JOIN authentication.tbl_user u2
                                ON mt.fk_last_modified_user_id =  u2.pk_user_id
                            LEFT JOIN authentication.tbl_user u3
                                ON mt.fk_deleted_user_id =  u3.pk_user_id
                        WHERE mt.int_sys_action_id != -1  
                        ORDER BY ;
                END;
            $$ LANGUAGE plpgsql;
        

------------------------------------------------------------------------------------------

            CREATE OR REPLACE PROCEDURE sp_insert_or_update_package_details(jsnPackageDetails JSONB, jsnSqlRes OUT JSON) AS $$
                DECLARE
                    objSpOutRes TYP_SP_OUT_RES;
                    objFnCheckRes TYP_FN_CHECK_RES;
                    dtmCurrent TIMESTAMP WITH TIME ZONE;
                BEGIN
                    SELECT NOW() INTO dtmCurrent;

                    -- // Duplication Checking in DB FUCTION(Insert and Update)
                    SELECT * FROM  fn_check_data_uniqueness('package.tbl_package_details'::VARCHAR, ''::VARCHAR, ::VARCHAR, 'pk_package_details_id'::VARCHAR, CAST(jsnPackageDetails->>'binPackageDetailsId' AS BIGINT)::BIGINT) INTO objFnCheckRes;
                    IF objFnCheckRes.sin_sql_check_status = -1::SMALLINT THEN
                        objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                        objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                        objSpOutRes.strSqlCustomMessage := objFnCheckRes.vhr_sql_check_message;

                        objSpOutRes.strSqlCustomErrorType := 'DATABASE_VALIDATION_ERROR';

                        SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                        RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                        RETURN;
                    END IF;

                    IF CAST(jsnPackageDetails->>'binPackageDetailsId' AS BIGINT)::BIGINT = 0::BIGINT THEN
                        -- // Count Exced or other checking License related in DB FUCTION(Insert)
                        

                        --jsnPackageDetails.intSysActionId = 0;
                        --jsnPackageDetails.binCreatedUserId = jsnPackageDetails.linLoginUserId;
                        --jsnPackageDetails.dtmCreated = jsnPackageDetails.dtmCurrentDateTime; --SELECT NOW();

                        INSERT INTO package.tbl_package_details (fk_package_master_id,
						sin_type,

						fk_commission_ac_id,
						fk_sup_fee_ac_id,
						fk_supplier_ac_id,
						fk_service_fee_ac_id,
						fk_discount_ac_id,

						txt_particulars,

						dbl_fare,
						dbl_tax,
						dbl_commission,
						dbl_sup_fee,
						dbl_cost,
						dbl_service_fee,
						dbl_discount,
						dbl_price)
                            VALUES(CAST(jsnPackageDetails->>'binPackageMasterId' AS BIGINT),
						CAST(jsnPackageDetails->>'sinType' AS SMALLINT),

						CAST(jsnPackageDetails->>'binCommissionAcId' AS BIGINT),
						CAST(jsnPackageDetails->>'binSupFeeAcId' AS BIGINT),
						CAST(jsnPackageDetails->>'binSupplierAcId' AS BIGINT),
						CAST(jsnPackageDetails->>'binServiceFeeAcId' AS BIGINT),
						CAST(jsnPackageDetails->>'binDiscountAcId' AS BIGINT),

						CAST(jsnPackageDetails->>'strParticulars' AS TEXT),

						CAST(jsnPackageDetails->>'dblFare' AS DOUBLE PRECISION),
						CAST(jsnPackageDetails->>'dblTax' AS DOUBLE PRECISION),
						CAST(jsnPackageDetails->>'dblCommission' AS DOUBLE PRECISION),
						CAST(jsnPackageDetails->>'dblSupFee' AS DOUBLE PRECISION),
						CAST(jsnPackageDetails->>'dblCost' AS DOUBLE PRECISION),
						CAST(jsnPackageDetails->>'dblServiceFee' AS DOUBLE PRECISION),
						CAST(jsnPackageDetails->>'dblDiscount' AS DOUBLE PRECISION),
						CAST(jsnPackageDetails->>'dblPrice' AS DOUBLE PRECISION)) RETURNING pk_package_details_id INTO objSpOutRes.binSqlPk;
                    ELSE
                        -- // UPDATE Check already update another usersin DB FUCTION(Update and Delete)
                        SELECT * FROM  fn_check_already_update_or_delete_another_user('package.tbl_package_details'::VARCHAR, 'pk_package_details_id'::VARCHAR, CAST(jsnPackageDetails->>'binPackageDetailsId' AS BIGINT)::BIGINT, CAST(jsnPackageDetails->>'intSysActionId' AS INT)::INT) INTO objFnCheckRes;
                        IF objFnCheckRes.sin_sql_check_status = -1::SMALLINT THEN
                            objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                            objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                            objSpOutRes.strSqlCustomMessage := objFnCheckRes.vhr_sql_check_message;

                            objSpOutRes.strSqlCustomErrorType := 'DATABASE_VALIDATION_ERROR';

                            SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                            RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                            RETURN;
                        END IF;

                        --jsnPackageDetails.intSysActionId = jsnPackageDetails.intSysActionId + 1;
                        --jsnPackageDetails.binLastModifiedUserId = jsnPackageDetails.linLoginUserId;
                        --jsnPackageDetails.dtmLastModified = jsnPackageDetails.dtmCurrentDateTime; --SELECT NOW();

                        UPDATE package.tbl_package_details SET fk_package_master_id = CAST(jsnPackageDetails->>'binPackageMasterId' AS BIGINT),
						sin_type = CAST(jsnPackageDetails->>'sinType' AS SMALLINT),

						fk_commission_ac_id = CAST(jsnPackageDetails->>'binCommissionAcId' AS BIGINT),
						fk_sup_fee_ac_id = CAST(jsnPackageDetails->>'binSupFeeAcId' AS BIGINT),
						fk_supplier_ac_id = CAST(jsnPackageDetails->>'binSupplierAcId' AS BIGINT),
						fk_service_fee_ac_id = CAST(jsnPackageDetails->>'binServiceFeeAcId' AS BIGINT),
						fk_discount_ac_id = CAST(jsnPackageDetails->>'binDiscountAcId' AS BIGINT),

						txt_particulars = jsnPackageDetails->>'strParticulars',

						dbl_fare = CAST(jsnPackageDetails->>'dblFare' AS DOUBLE PRECISION),
						dbl_tax = CAST(jsnPackageDetails->>'dblTax' AS DOUBLE PRECISION),
						dbl_commission = CAST(jsnPackageDetails->>'dblCommission' AS DOUBLE PRECISION),
						dbl_sup_fee = CAST(jsnPackageDetails->>'dblSupFee' AS DOUBLE PRECISION),
						dbl_cost = CAST(jsnPackageDetails->>'dblCost' AS DOUBLE PRECISION),
						dbl_service_fee = CAST(jsnPackageDetails->>'dblServiceFee' AS DOUBLE PRECISION),
						dbl_discount = CAST(jsnPackageDetails->>'dblDiscount' AS DOUBLE PRECISION),
						dbl_price = CAST(jsnPackageDetails->>'dblPrice' AS DOUBLE PRECISION) 
                        WHERE pk_package_details_id = CAST(jsnPackageDetails->>'binPackageDetailsId' AS BIGINT);
                    END IF;

                    --COMMIT;
                    objSpOutRes.sinSqlCustomStatus := 1::SMALLINT;
                    objSpOutRes.strSqlCustomStatusCode := 'SUCCESS';
                    objSpOutRes.strSqlCustomMessage := 'PackageDetails Saved';
                    objSpOutRes.dtmCurrent = dtmCurrent;
                    
                    SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                    RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                    RETURN;
                EXCEPTION WHEN OTHERS THEN
                    ROLLBACK;
                    objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                    objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                    objSpOutRes.strSqlCustomMessage := 'PackageDetails Save Failed';

                    objSpOutRes.strSqlCustomErrorType := 'DATABASE_SYSTEM_ERROR';

                    GET STACKED DIAGNOSTICS 
				        objSpOutRes.strSqlSysMessage := MESSAGE_TEXT,
                    	objSpOutRes.strSqlSysExceptionDetail := PG_EXCEPTION_DETAIL,
                    	objSpOutRes.strSqlSysExceptionHint := PG_EXCEPTION_CONTEXT;
                
                    SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                    RAISE EXCEPTION 'jsnSqlRes: %', jsnSqlRes;
                    RETURN;
                END;
            $$ LANGUAGE plpgsql;
        

------------------------------------------------------------------------------------------

            CREATE OR REPLACE PROCEDURE sp_delete_package_details(binPackageDetailsId BIGINT, intSysActionId INT, binDeletedUserId BIGINT, jsnSqlRes OUT JSON) AS $$
                DECLARE
                    objSpOutRes TYP_SP_OUT_RES;
                    objFnCheckRes TYP_FN_CHECK_RES;
                    dtmCurrent TIMESTAMP WITH TIME ZONE;
                BEGIN
                    SELECT NOW() INTO dtmCurrent;

                    -- // FOREIGN KEY REFERENCES Checking in DB FUCTION(Delete)
                    SELECT * FROM  fn_check_foreignkey_referance('tbl_package_details'::VARCHAR, 'pk_package_details_id'::VARCHAR, binPackageDetailsId::BIGINT, '{}'::VARCHAR ARRAY) INTO objFnCheckRes;
                    IF objFnCheckRes.sin_sql_check_status = -1::SMALLINT THEN
                        objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                        objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                        objSpOutRes.strSqlCustomMessage := objFnCheckRes.vhr_sql_check_message;

                        objSpOutRes.strSqlCustomErrorType := 'DATABASE_VALIDATION_ERROR';

                        SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                        RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                        RETURN;
                    END IF;

                    -- // UPDATE Check already update another users in DB FUCTION(Update and Delete)
                    SELECT * FROM  fn_check_already_update_or_delete_another_user('package.tbl_package_details'::VARCHAR, 'pk_package_details_id'::VARCHAR, CAST(binPackageDetailsId AS BIGINT)::BIGINT, intSysActionId::INT) INTO objFnCheckRes;
                    IF objFnCheckRes.sin_sql_check_status = -1::SMALLINT THEN
                        objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                        objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                        objSpOutRes.strSqlCustomMessage := objFnCheckRes.vhr_sql_check_message;

                        objSpOutRes.strSqlCustomErrorType := 'DATABASE_VALIDATION_ERROR';

                        SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                        RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                        RETURN;
                    END IF;

                    --dtmDeleted = SELECT NOW();
                    UPDATE package.tbl_package_details SET int_sys_action_id = -1::INT,
                                fk_deleted_user_id = binDeletedUserId,
                                dtm_deleted = dtmCurrent
                        WHERE pk_package_details_id = binPackageDetailsId;

                    --COMMIT;
                    objSpOutRes.sinSqlCustomStatus := 1::SMALLINT;
                    objSpOutRes.strSqlCustomStatusCode := 'SUCCESS';
                    objSpOutRes.strSqlCustomMessage := 'PackageDetails Deleted';
                    objSpOutRes.dtmCurrent = dtmCurrent;

                    SELECT * FROM fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                    RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                    RETURN;
                EXCEPTION WHEN OTHERS THEN
                    ROLLBACK;
                    objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                    objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                    objSpOutRes.strSqlCustomMessage := 'PackageDetails Delete Failed';

                    objSpOutRes.strSqlCustomErrorType := 'DATABASE_SYSTEM_ERROR';

                    GET STACKED DIAGNOSTICS 
				        objSpOutRes.strSqlSysMessage := MESSAGE_TEXT,
                    	objSpOutRes.strSqlSysExceptionDetail := PG_EXCEPTION_DETAIL,
                    	objSpOutRes.strSqlSysExceptionHint := PG_EXCEPTION_CONTEXT;

                    SELECT * FROM fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                    RAISE EXCEPTION 'jsnSqlRes: %', jsnSqlRes;
                    RETURN;
                END;
            $$ LANGUAGE plpgsql;
        

